#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Dr. Yves-Laurent Kom Samo"
__copyright__ = "Copyright (C) 2020 KXY Technologies, Inc."
__license__ = "GPL v3.0"
__version__ = "0.0"

from kxy.api import *
from kxy.regression import *
from kxy.classification import *
from kxy.data import *


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from kxy.api.core import least_continuous_mutual_information


def regression_difficulty(x, y):
	"""
	Estimates how difficult a regression problem is, using as 
	metric the inverse of the mutual information between features and output, :math:`1/I\\left(x, y\\right)`.

	The lower the mutual information (the higher its inverse), the more difficult
	it will be to predict :math:`y` using :math:`x`. The worst case scenario occurs 
	when :math:`y` and :math:`x` are statistically independent, in which case the difficulty
	is infinite. 

	The best case scenario occurs when :math:`y` is fully determined by :math:`x`,
	in which case the mutual information can be as high as :math:`+\\infty`, and the difficulty
	(its inverse) as low as 0.

	Parameters
	----------
	x : (n, d) np.array
		n i.i.d. draws from the features generating distribution.
	y : (n,) np.array
		n i.i.d. draws from the (continuous) laels generating distribution, sampled
		jointly with x.

	Returns
	-------
	d : float (non-negative)
		The regression difficulty.


	.. seealso:: 

		:ref:`kxy.api.core.mutual_information.least_continuous_mutual_information <least-continuous-mutual-information>`
	"""
	return 1./least_continuous_mutual_information(x, y)



def regression_feasibility(x, y):
	"""
	.. _regression-feasibility:
	Estimates how feasible a regression problem is, using as 
	metric the mutual information between features and output, :math:`I\\left(x, y\\right)`.

	The lower the mutual information, the more difficult it will be to predict :math:`y` using :math:`x`. 
	The worst case scenario occurs when :math:`y` and :math:`x` are statistically independent, 
	in which case the feasibility is 0. 

	The best case scenario occurs when :math:`y` is fully determined by :math:`x`,
	in which case the mutual information can be as high as :math:`+\\infty`.

	Parameters
	----------
	x : (n, d) np.array
		n i.i.d. draws from the features generating distribution.
	y : (n,) np.array
		n i.i.d. draws from the (continuous) laels generating distribution, sampled
		jointly with x.

	Returns
	-------
	d : float (non-negative)
		The regression feasibility.


	.. seealso:: 

		:ref:`kxy.api.core.mutual_information.least_continuous_mutual_information <least-continuous-mutual-information>`
	"""
	return least_continuous_mutual_information(x, y)





#!/usr/bin/env python
# -*- coding: utf-8 -*-

from kxy.api.core import least_mixed_mutual_information, discrete_mutual_information


def classification_difficulty(x_c, y, x_d=None):
	"""
	Estimates how difficult a classification problem is, using as 
	metric the inverse of the mutual information between features and output.

	.. math::
		\\frac{1}{I\\left({x_c, x_d}, y\\right)}

	Parameters
	----------
	x_c : (n, d) array_like
		n i.i.d. draws from the continuous data generating distribution.
	x_d : (n,) array_like or None (default)
		n i.i.d. draws from the discrete data generating distribution, jointly sampled with x_c, or None
		if there are no discrete features.
	y : (n,) array_like
		n i.i.d. draws from the (discrete) labels generating distribution, sampled jointly with x.

	Returns
	-------
	d : float
		The classification difficulty.


	.. seealso::  

		:ref:`kxy.api.core.mutual_information.least_mixed_mutual_information <least-mixed-mutual-information>`.
	"""
	return 1./least_mixed_mutual_information(x_c, y, x_d=x_d)



def classification_feasibility(x_c, y, x_d=None):
	"""
	.. _classification-feasibility:
	Estimates how feasible a classification problem is, using as 
	metric the mutual information between features and output.

	.. math::
		I\\left({x_c, x_d}, y\\right)


	Parameters
	----------
	x_c : (n, d) np.array or None
		n i.i.d. draws from the continuous data generating distribution, or None if there no continuous
		features. x_c and x_d cannot be both None.
	x_d : (n,) np.array or None (default)
		n i.i.d. draws from the discrete data generating distribution, jointly sampled with x_c, or None
		if there are no discrete features.
	y : (n,) np.array
		n i.i.d. draws from the (discrete) labels generating distribution, sampled jointly with x_c and/or x_d.


	Raises
	------
	AssertionError
		If x_d and x_c are both None.

	Returns
	-------
	d : float
		The classification feasibility.


	.. seealso:: 

		* :ref:`kxy.api.core.mutual_information.least_mixed_mutual_information <least-mixed-mutual-information>` 
		* :ref:`kxy.api.core.mutual_information.discrete_mutual_information <discrete-mutual-information>`
	"""
	assert x_d is not None or x_c is not None, "x_c and x_d cannot be both None."

	if x_c is None:
		return discrete_mutual_information(x_d, y)

	return least_mixed_mutual_information(x_c, y, x_d=x_d)





#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This file leverages the KXY API to solve various maximum entropy copula problems
under concordance measures. All optimization problems are solved by the KXY API
and users require an API key that should be set in the environment variable
KXY_API_KEY.
"""
import logging
import requests

import numpy as np
import scipy.special as spe

from kxy.api import APIClient
from .utils import avg_pairwise_spearman_corr, pre_conditioner

def max_ent_copula_entropy(method='average-pairwise-spearman-rho', d=None, rho=None,  \
		rho_features=None, rho_full=None):
	"""
	.. _max-ent-copula-entropy:
	Solves a constrained least-informative copula problem and returns the differential 
	entropy of the least informative copula satisfying the input constraints.

	.. important::

		The optimimzation problem is solved on the KXY infrastructure, not locally, and an 
		API key is required to do so.


	.. note::

		Let :math:`x \in \mathbb{R}^d` be a continuous random variable with probability density
		function :math:`p`, copula density :math:`c` and copula-uniform representation :math:`u := (F_1(x_i), \dots , F_d(x_d))` where 
		:math:`F_i` is the cummulative density function of coordinate variable :math:`x_i`. The differential entropy of :math:`x`, defined as 

		.. math::
			h(x) := -\int_{\mathbb{R}^d} p(x) \log p(x) dx

		can be broken down as 

		.. math::
			h(x) = h(u) + \sum_{i=1}^d h(x_i).

		:math:`h(u)=h(c)` is also the entropy of the copula of :math:`x` . Note that :math:`\sum_{i=1}^d h(x_i)-h(x)=-h(u)`
		is the total correlation of :math:`x` and quantifies information redundancy between 
		coordinates of :math:`x`.

		Given a functional :math:`g` of :math:`c` that can estimated from samples of math:`x` ,
		this function solves the following optimization problem over the set of all copula densities :math:`\mathcal{Co}`

		.. math::

			\\max_{c \in \mathcal{Co}} h(c),

			\\text{subject to } g(c) = \\alpha


		Functionals that are currently supported are population versions of the average pairwise 
		Spearman's rank correlation among a subset of coordinates of x.

		.. math::
			g(c) = 12 \\left[ \\frac{2}{d(d-1)} \\sum_{j<j^\\prime} E \\left( u_j u_{j^\\prime} \\right) \\right]-3

		For sample estimates based on :math:`x`, see Eq. (3.3.3) in [1]_.


	Parameters
	----------
	method : str (default: 'average-pairwise-spearman-rho')
		The type of constraint to use for the optimization problem. It should be either
		'average-pairwise-spearman-rho' or 'average-pairwise-spearman-rho-1vd'.

		When method='average-pairwise-spearman-rho', the functional :math:`g` used maps 
		a copula to its population average pairwise Spearman's rank correlation.

		When method='average-pairwise-spearman-rho-1vd', the functional :math:`g` used maps 
		a copula to its population average pairwise Spearman's rank correlation and the 
		population average pairwise Spearman's rank correlation of all its coordinates except one.
	d : int
		Dimension of the random variable. Defaulted to None, but should always be set.
	rho : float
		Average pairwise Spearman's rank correlation accross the full random vector. Defaulted to None,
		but should be set if method='average-pairwise-spearman-rho'.
	rho_full : float
		Average pairwise Spearman's rank correlation accross the full random vector. Defaulted to None,
		but should be set if method='average-pairwise-spearman-rho-1vd'.
	rho_features : float
		Average pairwise Spearman's rank correlation accross (d-1) coordinates. Defaulted to None,
		but should be set if method='average-pairwise-spearman-rho-1vd'.	

	Returns
	-------
	h : float
		The entropy of the least informative copula consistent with the input constraints.

	Raises
	------
	AssertionError
		If any required parameter is missing.

	ValueError
		If the optimization problem is not feasible.

	Exception
		If the KXY API is not able to return an answer quickly enough. This typically means
		you should try again later.


	.. rubric:: References

	.. [1] Joe, H. Journal of multivariate analysis 35 (1), 12-30, 1990.
	"""
	assert method in ('average-pairwise-spearman-rho', 'average-pairwise-spearman-rho-1vd'), \
		'Method not supported.'

	api_response = None
	if method == 'average-pairwise-spearman-rho':
		assert d is not None, 'The dimension d should be provided.'
		assert rho is not None, 'The average pairwise Spearman rho should be provided'

		api_response = APIClient.route(\
			path='/core/dependence/copula/maximum-entropy/entropy/rv/full', method='GET', \
			d=d, rho=rho)
		logging.info('rho: %.4f, d:%d, api_response: %s' % (rho, d, api_response.json()))

	if method == 'average-pairwise-spearman-rho-1vd':
		assert d is not None, 'The d dimension of the feature space should be provided'
		assert rho_features is not None, \
			'The average pairwise Spearman rank correlation between features should be provided'
		assert rho_full is not None, \
			'The average pairwise Spearman rank correlation including features and the ouput should be provided.'

		api_response = APIClient.route(\
			path='/core/dependence/copula/maximum-entropy/entropy/rv/1vd', method='GET', \
			d=d, rho_features=rho_features, rho_full=rho_full)
		logging.info('rho_features: %.4f, rho_full: %.4f, d: %d, api_response: %s' % (\
			rho_features, rho_full, d, api_response.json()))
	

	if api_response.status_code == requests.codes.ok:
		return api_response.json()['entropy']

	if api_response.status_code == 403:
		raise ValueError(api_response.json().get('reason', api_response.json().get('message')))

	if api_response.status_code == 404:
		raise Exception(api_response.json().get('reason', api_response.json().get('message')))

	return None


def scalar_continuous_entropy(x):
	"""
	.. _scalar-continuous-entropy:
	Estimates the (differential) entropy of a continuous scalar random variable using the standard 1-spacing estimator (see [2]_ and [3]_):

	.. math::
		h(x) \\approx - \gamma(1) + \\frac{1}{n-1} \\sum_{i=1}^{n-1} \log \\left[ n \\left(x_{(i+1)} - x_{(i)} \\right) \\right],

	where :math:`x_{(i)}` is the i-th smallest sample, and :math:`\\gamma` is
	the digamma function. 

	
	.. note::

		A Gaussian noise with negligible standard deviation (1/10000-th of the input standard deviation) is added to the input
		for robustness.

	
	Parameters
	----------
	x : (n,) np.array
		i.i.d. samples from the distribution of interest.

	Returns
	-------
	h : float
		The (differential) entropy of the continuous scalar random variable of interest.

	Raises
	------
	AssertionError
		If the input has the wrong shape.
		

	.. rubric:: References

	.. [2] Kozachenko, L. F., and Nikolai N. Leonenko. "Sample estimate of the entropy of a random vector." 
		Problemy Peredachi Informatsii 23.2 (1987): 9-16.

	.. [3] Beirlant, J., Dudewicz, E.J., Györfi, L., van der Meulen, E.C. "Nonparametric entropy estimation: an overview." 
		International Journal of Mathematical and Statistical Sciences. 6 (1): 17–40. (1997) ISSN 1055-7490. 
	"""
	assert len(x.shape) == 1 or x.shape[1] == 1, 'x should be a one dimensional numpy array'

	noise_std = x.std()/10000.
	x += noise_std * np.random.randn(*x.shape)
	sorted_x = np.unique(x)
	n = sorted_x.shape[0]
	ent = np.sum(np.log(n*(sorted_x[1:]-sorted_x[:-1])))/n - spe.digamma(1)

	return ent


def discrete_entropy(x):
	"""
	.. _discrete-entropy:
	Estimates the (Shannon) entropy of a discrete random variable taking up to q distinct values 
	given n i.i.d samples.

	.. math::
		h(x) = - \sum_{i=1}^q p_i \log p_i

	Parameters
	----------
	x : (n,) np.array
		i.i.d. samples from the distribution of interest.

	Returns
	-------
	h : float
		The (differential) entropy of the discrete random variable of interest.

	Raises
	------
	AssertionError
		If the input has the wrong shape.

	"""
	assert len(x.shape) == 1, 'Only flat numpy arrays are supported'
	categories = list(set(list(x)))
	n = len(x)
	probas = np.array([1.*len(x[x==cat])/n for cat in categories])

	return -np.dot(np.log(probas), probas)


def least_structured_copula_entropy(x):
	"""
	.. _least-structured-copula-entropy:
	Estimates the entropy of the least informative copula model whose average pairwise
	Spearman rank correlation is the same as the sample estimate from the input
	array.

	.. note::
	
		This also corresponds to least amount of total correlation that is evidenced
		by the sample average pairwise Spearman's rank correlation.

	Parameters
	----------
	x : (n, d) np.array
		n i.i.d. draws from the data generating distribution.

	Returns
	-------
	h : float
		The (differential) entropy of the least structured copula 
		consistent with observed average pairwise Spearman's rank correlation. See also :ref:`max_ent_copula_entropy <max-ent-copula-entropy>`.
	"""
	if len(x.shape) == 1 or x.shape[1] == 1:
		# By convention, the copula-dual representation 
		# of a 1d random variable is the uniform[0, 1].
		return 0.0

	d = x.shape[1]
	rho = avg_pairwise_spearman_corr(x)
	h = max_ent_copula_entropy(method='average-pairwise-spearman-rho', d=d, rho=rho)

	return h


def least_structured_continuous_entropy(x):
	""" 
	.. _least-structured-continuous-entropy:
	Estimates the entropy of a continuous d-dimensional random variable under the 
	least structured assumption for its copula. When :math:`d>1`,

	.. math::
		h(x) = h(u) + \sum_{i=1}^d h(x_i),

	and :math:`h(u)` is estimated using :ref:`least_structured_copula_entropy <least-structured-copula-entropy>` and 
	:math:`h(x_i)` are estimated using :ref:`scalar_continuous_entropy <scalar-continuous-entropy>`. 


	Parameters
	----------
	x : (n, d) np.array
		n i.i.d. draws from the data generating distribution.

	Returns
	-------
	h : float
		The (differential) entropy of the data generating distribution, assuming 
		its copula is the least structured copula consistent with the observed 
		average pairwise Spearman's rank correlation. 

		By convention, when :math:`d=1`, this function is the same as :ref:`scalar_continuous_entropy. <scalar-continuous-entropy>`, 
		and returns 0 when :math:`n=1`.
	"""
	if x.shape[0] == 1:
		# By convention, the entropy of a single sample is 0.
		return 0.0

	if len(x.shape) == 1 or x.shape[1] == 1:
		return scalar_continuous_entropy(x)

	x = x - x.mean(axis=0)
	a, log_abs_deta = pre_conditioner(x.T)
	z = np.dot(a, x.T).T

	ch = least_structured_copula_entropy(z)
	ih = np.sum([scalar_continuous_entropy(z[:, i]) for i in range(z.shape[1])])

	return ih+ch-log_abs_det_a


def least_structured_mixed_entropy(x_c, x_d):
	"""
	.. _least-structured-mixed-entropy:
	Estimates the joint entropy :math:`h(x_c, x_d)`, where :math:`x_c` is continuous
	random vector and :math:`x_d` is a discrete random vector.

	.. note::

		We use the identities

		.. math::
			h(x, y) &= h(y) + h(x|y) \\

			        &= h(y) + E \\left[ h(x \\vert y=.) \\right]

		that are true when :math:`x` and :math:`y` are either both continuous or both discrete
		to extend the definition of the joint entropy to the case where one is continuous and
		the other discrete.

		Specifically,

		.. math::
			h(x_c, x_d) = h(x_d) + \sum_{j=1}^q \mathbb{P}(x_d=j) h\\left(x_c \\vert x_d=j \\right).

		:math:`h(x_d)` is estimated using :ref:`discrete_entropy <discrete-entropy>`, :math:`\mathbb{P}(x_d=j)` is estimated using frequencies, and
		:math:`h\\left(x_c \\vert x_d=j \\right)` is estimated using :ref:`least_structured_continuous_entropy <least-structured-continuous-entropy>`.


	Parameters
	----------
	x_c : (n, d) np.array
		n i.i.d. draws from the continuous data generating distribution.

	x_d : (n,) np.array
		n i.i.d. draws from the discrete data generating distribution, jointly sampled with x_c.

	Returns
	-------
	h : float
		The entropy of the least structured distribution as evidenced by average pairwise Spearman
		rank correlations.
	"""
	labels = x_d.copy() if len(x_d.shape) == 1 else np.array([','.join([str(_) for _ in row]) for row in x_d])
	categories = list(set(labels))

	n = labels.shape[0]
	probas = np.array([1.*len(labels[labels==cat])/n for cat in categories])
	h = -np.dot(probas, np.log(probas))
	h += np.sum([probas[i] * least_structured_continuous_entropy(x_c[labels==categories[i]]) for i in range(len(categories))])

	return h






#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np 
from numpy.dual import svd

def avg_pairwise_spearman_corr(x):
	"""
	Calculates the sample average pairwise Spearman rank correlation 
	between columns of the 2-dimensional input array `x` of shape (n, d) as

	.. math::

		\\frac{12}{n^2-1} \\left[ \\frac{2}{d(d-1)} \\sum_{j<j^\\prime} \\left[ \\left( \\frac{1}{n} \\sum_{i=1}^n R_{ij} R_{ij^\\prime} \\right) - \\frac{(n+1)^2}{4}\\right] \\right]

	where :math:`R_{ij}` is the rank of :math:`x_{ij}` among :math:`x_{1j} \\dots x_{nj}`. 

	See Eq. (3.3.3) in [1]_.


	Parameters
	----------
	x : (n, d) np.array
		Input data representing n i.i.d. draws from the d-dimensional 
		random variable, whose average pairwise Spearman rank correlation
		this function calculates.


	Returns
	-------
	rho : float
		The average pairwise Spearman rank correlation.



	.. rubric:: References

	.. [1] Joe, H. Journal of multivariate analysis 35 (1), 12-30, 1990.
	"""
	# R[i,j] is the rank of x[i, j] among x[1, j] ... x[n, j]
	n, m = x.shape
	R = 1+x.argsort(axis=0).argsort(axis=0)
	spearman_corr = np.cov(R.T, bias=True)/((n*n-1.)/12.)
	rho = (np.sum(spearman_corr)-m)/(m*(m-1.))

	return rho



def pre_conditioner(x):
	"""
	Given n i.i.d. samples from a random variable x, compute a square matrix A such that 
	coordinates of :math:`z = Ax` have similar pairwise Spearman rank correlation.

	.. note::

		The current implementation computes A as :math:`A = D^{-\\frac{1}{2}} U^T` where :math:`C = UDU^T` 
		is the SVD decomposition of the covariance matrix of :math:`x`.

		It is worth recalling that if :math:`x` is mean 0 and has covariance matrix :math:`C`, then 
		:math:`z := D^{-\\frac{1}{2}} U^T x` is also mean 0 and has covariance matrix the identity matrix.

		The Spearman rank correlation of :math:`z` might not be the identity matrix, but its off diagonal terms
		will have the same magnitude (fairly small).


	Parameters
	----------
	x : (n, d) np.array
		Input data representing n i.i.d. draws from the d-dimensional 
		random variable.


	Returns
	-------
	A : (d, d) np.array
		The pre-conditioning matrix.
	ld: float
		:math:`\\log |\\text{det} A|`
	"""
	cov = np.cov(x)
	u, s, v = svd(cov)
	eps = 0.0
	oc = np.max(s)/np.min(s)
	if oc > 1.0:
		nc = np.min([oc, 1e3])
		eps = np.min(s)*(oc-nc)/(nc-1.0)
	
	a = np.dot(np.diag(1.0/(np.sqrt(np.absolute(s) + eps))), u.T)
	det_a = np.prod(1.0/(np.sqrt(np.absolute(s) + eps)))

	return a, np.log(np.abs(det_a))





#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This file leverages the KXY API to evaluate mutual information and related measure derived
from solving maximum entropy copula problems under concordance measures. All optimization 
problems are solved by the KXY API and users require an API key  that should be set in the 
environment variable ``KXY_API_KEY``.
"""
import numpy as np

from .utils import avg_pairwise_spearman_corr, pre_conditioner
from .entropy import least_structured_copula_entropy, least_structured_continuous_entropy, \
	least_structured_mixed_entropy, max_ent_copula_entropy, discrete_entropy


def least_total_correlation(x):
	""" 
	.. _least-total-correlation:
	Estimates the smallest total correlation of a continuous d-dimensional random vector,
	that is consistent with the average pairwise Spearman rank correlation estimated 
	from the input sample.

	.. math::
		TC(x) := \sum_{i=1}^d h(x_i)-h(x) = -h(u).

	This is the negative of :ref:`kxy.api.core.entropy.least_structured_copula_entropy <least-structured-copula-entropy>`.

	Parameters
	----------
	x : (n, d) np.array
		n i.i.d. draws from the data generating distribution.

	Returns
	-------
	tc : float
		The smallest total correlation consistent with observed empirical evidence, in nats.

	Raises
	------
	AssertionError
		If x is not a two dimensional array.
	"""
	assert len(x.shape) == 2
	return -least_structured_copula_entropy(x)



def least_continuous_mutual_information(x,  y):
	"""
	.. _least-continuous-mutual-information:
	Estimates the mutual information between a d-dimensional random vector :math:`x` of features
	and a continuous scalar random variable :math:`y` (or label), assuming the least amount of 
	structure in :math:`x` and :math:`(x, y)`, other than what is necessary to be consistent with some 
	observed concordance measures. 

	.. note::

		.. math::
			I(x, y) &= h(x) + h(y) - h(x, y) \\

			        &= h\\left(u_x\\right) - h\\left(u_{x,y}\\right)
		
		where :math:`u_x` (resp. :math:`u_{x,y}`) is the copula-uniform representation of :math:`x`
		(resp. :math:`(x, y)`).

		We use as model for :math:`u_x` the least structured copula that is consistent with with the
		average pairwise Spearman rank correlation between coordinates of :math:`x`. 

		We use as model for :math:`u_{x, y}` the least structured copula that is consistent with both 
		the average pairwise Spearman rank correlation between coordinates of :math:`x`, and the 
		average pairwise Spearman rank correlation between coordinates of :math:`(x, y)`.

	Parameters
	----------
	x : (n, d) np.array
		n i.i.d. draws from the features generating distribution.
	y : (n,) np.array
		n i.i.d. draws from the (continuous) laels generating distribution, sampled
		jointly with x.

	Returns
	-------
	i : float
		The mutual information between x and y, in nats.

	Raises
	------
	AssertionError
		If x is not a two dimensional array.
	"""
	assert np.can_cast(x, float) and np.can_cast(y,  float), 'x and y should represent draws from continuous random variables.'
	assert len(y.shape) == 1 or y.shape[1] == 1, 'Only one-dimensional outputs are supported for now.'

	x_ = np.reshape(x, (len(x), 1)) if len(x.shape) == 1 else x.copy()
	y_ = np.reshape(y, (len(y), 1)) if len(y.shape) == 1 else y.copy()

	if len(x.shape) == 1 or x.shape[1] == 1:
		return least_total_correlation(np.hstack([x_, y_]))

	x = x - x.mean(axis=0)
	a, log_abs_deta = pre_conditioner(x.T)
	z = np.dot(a, x.T).T

	rho_features = avg_pairwise_spearman_corr(z)
	rho_full = avg_pairwise_spearman_corr(np.hstack([z, y_]))
	d = z.shape[1]

	hux = max_ent_copula_entropy(method='average-pairwise-spearman-rho', \
		rho=rho_features, d=d)

	huxy = max_ent_copula_entropy(method='average-pairwise-spearman-rho-1vd', \
		rho_full=rho_full, rho_features=rho_features, d=d)

	return max(hux-huxy, 0.0)



def least_mixed_mutual_information(x_c, y, x_d=None):
	"""
	.. _least-mixed-mutual-information:
	Estimates the mutual inforrmation between some features (a d-dimensional random vector
	:math:`x_c` and possibly a discrete feature variable :math:`x_d` and a discrete output/label 
	random variable.	

	.. note::

		.. math::
			I({x_c, x_d}, y) &= h(y) + h(x_c, x_d) - h(x_c, x_d, y) \\

							 &= h(x_c, x_d) - h\\left(x_c, x_d \\vert y\\right) \\

							 &= h(x_c, x_d) - E\\left[h\\left(x_c, x_d \\vert y=.\\right)\\right] \\

							:&= h(x_c, x_d) - \sum_{j=1}^q \mathbb{P}(y=j)  h\\left(x_c, x_d \\vert y=j \\right)

		
		When there are discrete features, :math:`h(x_c, x_d)` is estimated using :ref:`kxy.api.core.entropy.least_structured_mixed_entropy <least-structured-mixed-entropy>`,
		:math:`\mathbb{P}(y=j)` is estimated using frequencies, :math:`h\\left(x_c, x_d | y=j\\right)` is estimated 
		using :ref:`kxy.api.core.entropy.least_structured_mixed_entropy <least-structured-mixed-entropy>`.

		When there are no discrete features, :math:`x_d` can simply be removed from the equations above, and 
		:ref:`kxy.api.core.entropy.least_structured_continuous_entropy <least-structured-continuous-entropy>` is used for estimation instead of 
		:ref:`kxy.api.core.entropy.least_structured_mixed_entropy <least-structured-mixed-entropy>`.

	Parameters
	----------
	x_c : (n, d) np.array
		n i.i.d. draws from the continuous data generating distribution.
	x_d : (n,) np.array or None (default)
		n i.i.d. draws from the discrete data generating distribution, jointly sampled with x_c, or None
		if there are no discrete features.
	y : (n,) np.array
		n i.i.d. draws from the (discrete) labels generating distribution, sampled jointly with x.

	Returns
	-------
	i : float
		The mutual information between features and discrete labels, in nats.

	Raises
	------
	AssertionError
		If y is not a one-dimensional array, or x_c is not an array of numbers.
	"""
	assert np.can_cast(x_c, float), 'x_c should represent draws from a continuous random variable.'
	assert len(y.shape) == 1 or y.shape[1] == 1, 'Only one-dimensional outputs are supported for now.'

	categories = list(set(list(y)))
	n = y.shape[0]
	probas = np.array([1.*len(y[y==cat])/n for cat in categories])

	h = least_structured_continuous_entropy(x_c) if x_d is None else least_structured_mixed_entropy(x_c, x_d)
	wh = np.sum([probas[i] * least_structured_continuous_entropy(x_c[y==categories[i]]) for i in range(len(categories))]) if x_d is None \
		else np.sum([probas[i] * least_structured_mixed_entropy(x_c[y==categories[i]], x_d[y==categories[i]]) for i in range(len(categories))])

	return max(h-wh, 0.0)



def discrete_mutual_information(x, y):
	"""
	.. _discrete-mutual-information:
	Estimates the (Shannon) mutual information between two discrete random variables from
	i.i.d. samples.

	Parameters
	----------
	x : (n,) np.array or None (default)
		n i.i.d. draws from a discrete distribution.
	y : (n,) np.array
		n i.i.d. draws from another discrete distribution, sampled jointly with x.

	Returns
	-------
	i : float
		The mutual information between x and y, in nats.

	Raises
	------
	AssertionError
		If y or x is not a one-dimensional array, or if x and y have different shapes.
	"""
	assert len(y.shape) == 1 or y.shape[1] == 1, 'Only one-dimensional outputs are supported for now.'
	assert len(x.shape) == 1 or x.shape[1] == 1, 'Only one-dimensional outputs are supported for now.'
	assert x.shape[0] == y.shape[0], 'Both arrays should have the same dimension.'

	hx = discrete_entropy(x)
	hy = discrete_entropy(y)
	hxy = discrete_entropy(np.array([str(x[i]) + '*_*' + str(y[i]) for i in range(x.shape[0])]))

	return max(0., hx+hy-hxy)





#!/usr/bin/env python
# -*- coding: utf-8 -*-

from kxy.api.core import least_continuous_mutual_information


def regression_suboptimality(yp, y, x):
	"""
	.. _regression-suboptimality:
	Quantifies the extent to which a regression model can be improved without requiring adding features.

	.. note::

		The aim of a regression model is to find the function :math:`x \\to f(x) \in \\mathbb{R}` 
		to be used as our predictor for :math:`y` given :math:`x` so that :math:`f(x)` fully captures all
		the information about :math:`y` that is contained in :math:`x`.  For instance, this will be 
		the case when the true generative model takes the form

		.. math::

			y = f(x) + \\epsilon

		with :math:`\\epsilon` statistically independent from :math:`y`, in which case :math:`h \\left( y \\vert x \\right) = h(\\epsilon)`.

		More generally, the conditional entropy :math:`h \\left( y \\vert x \\right)` represents 
		the amount of information about :math:`y` that cannot be explained by :math:`x`, while 
		:math:`h \\left( y \\vert \\tilde{f}(x) \\right)` represents the amount of information 
		about :math:`y` that cannot be explained by the regression model 

		.. math::

			y = \\tilde{f}(x) + \\epsilon.

		A natural metric for how suboptimal a particular regression model is can therefore be defined as
		the difference between the amount of information about :math:`y` that cannot be explained by 
		:math:`\\tilde{f}(x)` and the amount of information about :math:`y` that cannot be explained by :math:`x`


		.. math::

			\\text{subopt}(\\tilde{f}; x) &= h \\left( y \\vert \\tilde{f}(x) \\right) - h \\left( y \\vert x \\right) \\

			:&= I\\left(y, x \\right) - I\\left(y, \\tilde{f}(x) \\right) \\

			&\geq 0.

		This regression suboptimality metric is 0 if and only if :math:`\\tilde{f}(x)` fully captures any information about :math:`y`
		that is contained in :math:`x`. When 

		.. math::

			\\text{subopt}(\\tilde{f}; x) > 0 

		on the other hand, there exists a regression model using :math:`x` as features that can better predict :math:`y`. The larger 
		:math:`\\text{subopt}(\\tilde{f}; x)`, the more the regression model is suboptimal and can be improved.


	Parameters
	----------
	x : (n, d) np.array
		n i.i.d. draws from the features generating distribution.
	y : (n,) np.array
		n i.i.d. draws from the (continuous) labels generating distribution, sampled
		jointly with x.
	yp : (n,) np.array
		Predictions of y.

	Returns
	-------
	d : float
		The regression's suboptimality measure.


	.. seealso:: 

		:ref:`kxy.api.core.mutual_information.least_continuous_mutual_information <least-continuous-mutual-information>`
	"""
	return least_continuous_mutual_information(x, y)-least_continuous_mutual_information(yp, y)




def regression_additive_suboptimality(e, x):
	"""
	.. _regression-additive-suboptimality:
	Quantifies the extent to which a regression model can be improved without requiring additional features, by evaluating 
	how informative its residuals still are about the features.

	.. note::

		Additive regression models aim at breaking down a label :math:`y` as the sum of a component that solely depend on 
		features :math:`f(x)` and a residual component that is statistically independent from features :math:`\\epsilon`

		.. math::

			y = f(x) + \\epsilon.

		In an ideal scenario, the regreession residual :math:`\\epsilon` would indeed be stastically independent from the features
		:math:`x`. In pratice however, this might not be the case, for instance when the space of candidate functions used by
		the regression model isn't flexible enough (e.g. linear regression or basis functions regression), or the optimization
		has not converged to the global optimum. 

		Any departure from statistical independence between residuals :math:`\\epsilon` and features :math:`x` is an indication that what
		:math:`x` can reveal about :math:`y` is not fully captured by :math:`f(x)`, which implies that the regression model can be improved.

		Thus, we define the additive suboptimality of a regression model as the mutual information between its residuals and its features

		.. math::

			\\text{add-subopt}(\\tilde{f}; x) := I\\left( y-\\tilde{f}(x), x \\right)


	Parameters
	----------
	e : (n,) np.array
		Regression residuals.
	x : (n,) np.array
		Regression features.


	Returns
	-------
	d : float
		The regression's additive suboptimality measure.


	.. seealso:: 

		:ref:`kxy.api.core.mutual_information.least_continuous_mutual_information <least-continuous-mutual-information>`
	"""
	return least_continuous_mutual_information(x, e)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This file leverages the KXY API to evaluate mutual information rate and related measure derived
from solving mmaximum entropy copula problems under concordance measures.  All optimization problems 
are solved by the KXY API and users require an API key that should be set in the environment variable
KXY_API_KEY.
"""


# -*- coding: utf-8 -*-
"""
Created on Thu Feb 27 10:54:00 2020

@author: ylkomsamo
"""

import sys
sys.path.append('.')
from setuptools import setup, find_packages

setup(name="kxy",
	version="0.0.6",
	zip_safe=False,
	license="AGPLv3",
    author="Dr. Yves-Laurent Kom Samo",
    author_email="github@kxy.ai",
	url="https://docs.kxysolutions.com",
	description = "Python API to the AutoML KXY platform", 
	long_description="The KXY platform is the first and only AutoML platform for model-free" \
		" pre-learning and post-learning. More information at https://docs.kxysolutions.com.",
    project_urls={
        "Documentation": "https://docs.kxysolutions.com",
        "Source Code": "https://github.com/kxytechnologies/kxy-python/"},
	download_url = "https://github.com/kxytechnologies/kxy-python/archive/v0.0.3.tar.gz",
	keywords = ["AutoML", "Pre-Learning", "Post-Learning", "Model-Free ML"],
	packages=find_packages(exclude=["tests"]),
	install_requires=["numpy>=1.13.1", "pandas>=0.19.2", "pandas>=0.19.2", "requests==2.22.0"],
    classifiers=[
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Programming Language :: Python :: 3 :: Only",
        "Development Status :: 3 - Alpha",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Mathematics"
    ]
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This file leverages the KXY API to solve various maximum entropy rate copula problems
under concordance measures. All optimization problems are solved by the KXY API
and users require an API key that should be set in the environment variable ``KXY_API_KEY``.
"""